import os, sys, struct, re
from distutils import sysconfig, util
from distutils.ccompiler import new_compiler, show_compilers
from distutils.core import Command
from distutils.dep_util import newer, newer_group

from Structures import Script, ExecutableScript, Executable

STUB_MAIN_SOURCE = os.path.join(os.path.dirname(__file__), 'stubmain.c')

SHELL_SCRIPT = """\
#!/usr/bin/env python

# Force Python to ignore the current directory when loading the
# first component of the import path
import sys, imp
module_info = imp.find_module('%(base_module)s', filter(None, sys.path))
imp.load_module('%(base_module)s', *module_info)

%(import_cmd)s
%(classname)s().run()
"""

class BuildScripts(Command):

    command_name = 'build_scripts'

    description = "\"build\" scripts"

    user_options = [
        ('build-dir=', 'd', "directory to \"build\" (copy) to"),
        ('build-temp=', 't',
         "directory for temporary files (build by-products)"),
        ('force', 'f', "forcibly build everything (ignore file timestamps"),
        ('debug', 'g', "compile/link with debugging information"),
        ('compiler=', 'c', "specify the compiler type"),
        ]

    help_options = [
        ('help-compiler', None,
         "list available compilers", show_compilers),
        ]

    boolean_options = ['force', 'debug']

    stub_name = 'stubmain'

    def initialize_options(self):
        self.build_dir = None
        self.scripts = None
        self.force = None
        self.outfiles = None
        self.build_temp = None
        self.compiler = None
        self.debug = None
        self.link_preargs = []
        self.link_postargs = []
        self.exe_extension = None
        self.stub_filename = None
        return
    
    def finalize_options(self):
        undefined_temp = self.build_temp is None

        self.set_undefined_options('build',
                                   ('build_scripts', 'build_dir'),
                                   ('build_temp', 'build_temp'),
                                   ('compiler', 'compiler'),
                                   ('debug', 'debug'),
                                   ('force', 'force'))

        self.scripts = self.distribution.scripts
        self.outfiles = []

        if undefined_temp:
            self.build_temp = os.path.join(self.build_temp, 'scripts')

        self.exe_extension = sysconfig.get_config_var('EXE') or ''
        if os.name == 'nt' and self.debug:
            self.exe_extension = '_d' + self.exe_extension

        self.stub_filename = os.path.join(self.build_temp,
                                          self.stub_name + self.exe_extension)
        return

    def _prep_compiler(self):
        # Setup the CCompiler object that we'll use to do all the
        # compiling and linking
        self.compiler = new_compiler(compiler=self.compiler,
                                     verbose=self.verbose,
                                     dry_run=self.dry_run,
                                     force=self.force)
        sysconfig.customize_compiler(self.compiler)
        
        # Make sure Python's include directories (for Python.h, pyconfig.h,
        # etc.) are in the include search path.
        py_include = sysconfig.get_python_inc()
        plat_py_include = sysconfig.get_python_inc(plat_specific=1)
        include_dirs = [py_include]
        if plat_py_include != py_include:
            include_dirs.append(plat_py_include)
        self.compiler.set_include_dirs(include_dirs)

        if os.name == 'posix':
            # Add the Python archive library
            ldlibrary = sysconfig.get_config_var('BLDLIBRARY')
            # MacOSX with frameworks doesn't link against a library
            if ldlibrary:
                # Get the location of the library file
                for d in sysconfig.get_config_vars('LIBDIR', 'LIBP', 'LIBPL'):
                    library = os.path.join(d, ldlibrary)
                    if os.path.exists(library):
                        self.compiler.add_link_object(library)
                        break
                    

            args = sysconfig.get_config_vars('LDFLAGS', 'LINKFORSHARED')
            self.link_preargs = ' '.join(args).split()

            args = sysconfig.get_config_vars('LIBS', 'MODLIBS', 'SYSLIBS',
                                             'LDLAST')
            self.link_postargs = ' '.join(args).split()
        elif os.name == 'nt':
            # Add Python's library directory            
            lib_dir = os.path.join(sys.exec_prefix, 'libs')
            self.compiler.add_library_dir(lib_dir)
        return

    def _prep_stub(self):
        if self.force or newer(STUB_MAIN_SOURCE, self.stub_filename):
            # Create the executable stub
            self.announce("building stub executable")

            # Force rebuild of scripts
            self.force = 1
            
            if os.name == 'nt' and self.debug:
                executable = self.stub_name + '_d'
            else:
                executable = self.stub_name

            objects = self.compiler.compile([STUB_MAIN_SOURCE],
                                            output_dir=self.build_temp,
                                            debug=self.debug)

            self.compiler.link_executable(objects, executable,
                                          output_dir=self.build_temp,
                                          debug=self.debug,
                                          extra_preargs=self.link_preargs,
                                          extra_postargs=self.link_postargs)
        else:
            self.announce("skipping stub executable (up-to-date)")
        return

    _include_re = re.compile(r'^\s*#\s*include\s*("|<)(?P<file>[^">]+?)("|>)',
                             re.MULTILINE)

    def _find_includes(self, script, filename, includes):
        source = open(filename).read()
        basename = os.path.basename(filename)

        include_dirs = [os.path.dirname(filename)]
        include_dirs.extend(script.include_dirs)

        match = self._include_re.search(source)
        while match:
            filename = util.convert_path(match.group('file'))
            for base in include_dirs:
                include = os.path.normpath(os.path.join(base, filename))
                if os.path.isfile(include) and include not in includes:
                    includes.append(include)
                    self._find_includes(script, include, includes)
                    break
            match = self._include_re.search(source, match.end())
        return

    def _prep_build(self, script):
        # This should really exist in the CCompiler class, but
        # that would required overriding all compilers.
        result = []
        for source in script.sources:
            includes = []
            self._find_includes(script, util.convert_path(source), includes)
            result.append((source, includes))
        return result

    def run(self):
        """
        Create the proper script for the current platform.
        """
        if not self.scripts:
            return

        # Ensure the destination directory exists
        self.mkpath(self.build_dir)

        # Break out the various types of scripts
        scripts = [ x for x in self.scripts
                        if isinstance(x, Script) ]
        executables = [ x for x in self.scripts
                        if isinstance(x, Executable) ]
        exec_scripts = [ x for x in self.scripts
                         if isinstance(x, ExecutableScript) ]

        # Create the compiler for any Windows script, executables or
        # executable scripts
        if (os.name == 'nt' and (scripts or exec_scripts)) or executables:
            self._prep_compiler()

        # Create the executable stub for any Windows script or executable
        # scripts
        if os.name == 'nt' and (scripts or exec_scripts):
            self._prep_stub()

        for script in scripts:
            self.build_script(script)

        for executable in executables:
            self.build_executable(executable)

        for exec_script in exec_scripts:
            self.build_executable_script(exec_script)

        return

    def write_executable_script(self, outfile, script):
        # Get the stub executable data
        f = open(self.stub_filename, 'rb')
        stub_data = f.read()
        f.close()
        
        # append the data to the stub executable
        f = open(outfile, 'wb')

        # Write the executable
        f.write(stub_data)

        # Write the script_data struct
        # struct script_data {
        #   int tag;
        #   int length;
        # }
        script_data = struct.pack("ii", 0x1234567A, len(script))
        f.write(script_data)
                
        # Write the script string
        f.write(script)
                    
        # Write the script_info struct
        # struct script_info {
        #   int tag;
        #   int offset;
        # }
        script_info = struct.pack("ii", 0x1234567A, len(stub_data))
        f.write(script_info)

        f.close()
        

    def build_script(self, script):
        outfile = self.get_script_filename(script)

        # The "source" for normal scripts is created from this file
        if not (self.force or newer(__file__, outfile)):
            self.announce("skipping '%s' script (up-to-date)" % script.name)
            return
        else:
            self.announce("building '%s' script" % script.name)

        if not self.dry_run:
            args = {'classname' : script.identifier}
                
            # Calculate the base module for the script module
            index = script.path.find('.')
            if index == -1:
                args['base_module'] = script.path
            else:
                args['base_module'] = script.path[:index]

            if script.module:
                args['import_cmd'] = 'from %s import %s' % (script.module, 
                                                            script.identifier)
            else:
                args['import_cmd'] = 'import %s' % script.identifier

            script_data = SHELL_SCRIPT % args

            if os.name == 'nt':
                self.write_executable_script(outfile, script_data)
            else:
                f = open(outfile, 'w')
                f.write(script_data)
                f.close()
        return

    def build_executable_script(self, exec_script):
        outfile = self.get_script_filename(exec_script)

        source = util.convert_path(exec_script.source)
        sources = [__file__, self.stub_filename, source]

        if not (self.force or newer_group(sources, outfile, 'newer')):
            self.announce("skipping '%s' executable script (up-to-date)" %
                          exec_script.name)
            return
        else:
            self.announce("building '%s' executable script" % exec_script.name)

        # Get the __main__ Python code body
        f = open(source, 'r')
        script_data = f.read()
        f.close()
        
        if not self.dry_run:
            if os.name == 'nt':
                self.write_executable_script(outfile, script_data)
            else:
                f = open(outfile, 'w')
                f.write(script_data)
                f.close()
        return

    def build_executable(self, executable):
        outfile = self.get_script_filename(executable)

        all_sources = self._prep_build(script)
        sources = []
        for source, includes in all_sources:
            sources.append(source)
            sources.extend(includes)

        if not (self.force or newer_group(sources, outfile, 'newer')):
            self.announce("skipping '%s' executable (up-to-date)" %
                          executable.name)
            return
        else:
            self.announce("building '%s' executable" % executable.name)

        output_dir = os.path.join(self.build_temp, executable.name)
        
        macros = executable.define_macros[:]
        for undef in executable.undef_macros:
            macros.append((undef,))

        objects = []
        for source, includes in all_sources:
            if not self.force:
                # Recompile if the includes or source are newer than the
                # resulting object files.
                objs = self.compiler.object_filenames([source], 1, output_dir)
                
                # Recompile if any of the inputs are newer than the object
                inputs = [source] + includes
                force = 0
                for filename in objs:
                    force = force or newer_group(inputs, filename, 'newer')
                self.compiler.force = force

            objs = self.compiler.compile(
                [source],
                output_dir=output_dir,
                macros=macros,
                include_dirs=executable.include_dirs,
                debug=self.debug,
                extra_postargs=executable.extra_compile_args)
            objects.extend(objs)

        # Reset the force flag on the compilier
        self.compiler.force = self.force
        
        # Now link the object files together into a "shared object" --
        # of course, first we have to figure out all the other things
        # that go into the mix.
        if os.name == 'nt' and self.debug:
            executable = executable.name + '_d'
        else:
            executable = executable.name

        if executable.extra_objects:
            objects.extend(executable.extra_objects)

        self.compiler.link_executable(
            objects, executable,
            libraries=executable.libraries,
            library_dirs=executable.library_dirs,
            runtime_library_dirs=executable.runtime_library_dirs,
            extra_preargs=self.link_preargs,
            extra_postargs=self.link_postargs + executable.extra_link_args,
            debug=self.debug,
            build_temp=self.build_temp)
        
        return

    def get_script_filename(self, script):
        """
        Convert the name of a script into the name of the file which it
        will be run from.
        """
        # All Windows scripts are executables
        if os.name == 'nt' or (isinstance(script, Executable) or 
                               isinstance(script, ExecutableScript)):
            script_name = script.name + self.exe_extension
        else:
            script_name = script.name
        return os.path.join(self.build_dir, script_name)
        
    def get_outputs(self):
        outfiles = []
        for script in self.scripts:
            outfiles.append(self.get_script_filename(script))
        return outfiles

    def get_source_files(self):
        filenames = []
        for script in self.scripts:
            if isinstance(script, ExecutableScript):
                filenames.append(script.source)
        return filenames