########################################################################
#
# File Name:            CoreFunctions.py
#
#
"""

Definition of Core Versa Functions

WWW: http://4suite.org/4RDF        e-mail: support@4suite.org

Copyright 2002 Fourthought Inc, USA.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import types
import DataTypes
from Ft.Rdf import RDF_MS_BASE, RDF_SCHEMA_BASE
from Ft.Rdf.Parsers.Versa import Constants
from Ft.Rdf.Parsers.Versa import Compile
from Ft.Rdf.Parsers.Versa import RDF_TYPE_PROP, RDF_SUBCLASSOF_PROP, RDF_SUBPROPOF_PROP
import ResourceExpressions, Traversal, Literals, Optimizer
from Ft.Lib import boolean, number

_TYPE = RDF_MS_BASE + 'type'
_SUBCLASSOF = RDF_SCHEMA_BASE + 'subClassOf'


def Current(con):
    """
    Return the current value
    """
    return con.current


#
# List and set functions
#

def List(con, *args):
    """
    Create a new list
    """
    return list(args[:])


def Set(con, *args):
    """
    Create a new set.
    If the argument is a single list, it is converted to a similar set
    If the argument is one or more scalar value, it is converted to a set with those values
    Otherwise, it is an error, and you may have meant to use the flatten() function
    """
    if isinstance(args[0], types.ListType):
        return DataTypes.Set(args[0])
    else:
        return DataTypes.Set(list(args[:]))


def Length(con, obj):
    """
    Return the length of a list
    """
    return len(DataTypes.ToList(obj))


def Join(con, *args):
    """
    Create a new list
    """
    return reduce(lambda a, b: a + DataTypes.ToList(b), args)


def Item(con, l, num=1):
    """
    Converts the first argument to a list L, and the second to a number N.
    Returns the Nth item in L.  N defaults to 1.
    """
    index = DataTypes.ToNumber(num)
    try:
        return DataTypes.ToList(l)[index-1]
    except:
        return []


def Head(con, l, num=1):
    """
    Converts the first argument to a list L, and the second to a number N.
    Returns a list consisting of the first N items in L.  N defaults to 1.
    If N is negative, or exceeds the length of the list, the entire list is
    the result.
    """
    return DataTypes.ToList(l)[:DataTypes.ToNumber(num)]


def Tail(con, l, num=1):
    """
    Converts the first argument to a list L, and the second to a number N.
    Returns a list consisting of the last N items in L.  N defaults to 1.
    If N is negative, or exceeds the length of the list, an empty list is
    the result.
    """
    return DataTypes.ToList(l)[-DataTypes.ToNumber(num):]


def Rest(con, l, num=1):
    """
    Converts the first argument to a list L, and the second to a number N.
    Returns a list consisting of all items in L after position N.
    N defaults to 1.  If N is negative, or exceeds the length of the list,
    an empty list is the result.  The following expression returns the same
    list as L, regardless of the value of N:
    """
    return DataTypes.ToList(l)[DataTypes.ToNumber(num):]


#Instead of these forms, add a "flatten" function
##def List(con, *args):
##    """
##    Create a new list
##    """
##    finalValues = []
##    for arg in args:
##        if DataTypes.IsList(arg):
##            finalValues.extend(arg)
##        elif DataTypes.IsSet(arg):
##            finalValues.extend(arg._d.keys())
##        else: 
##            finalValues.append(arg)
##    return finalValues


##def Set(con, *args):
##    """
##    Create a new set
##    """
##    finalValues = []
##    for arg in args:
##        if DataTypes.IsList(arg):
##            finalValues.extend(arg)
##        elif DataTypes.IsSet(arg):
##            finalValues.extend(arg._d.keys())
##        else: 
##            finalValues.append(arg)
##    rt = DataTypes.Set(finalValues)
##    return rt


def Distribute(con, sourceList, *exprs):
    """
    Apply the query to each of the resources in the resource Expression.
    The first argument must be a list of resources
    The second is a list of queries.
    the results is a list of lists.  The nested list will can the result of each
    of the queries.
    """
    sq_funcs = [
        Optimizer.SubQueryFunction(
          Compile(DataTypes.ToString(e)), con
        )
        for e in exprs
        ]
    sourceList = DataTypes.ToList(sourceList)
    res = []
    orig = con.current
    for obj in sourceList:
        #FIXME: ShiftContext is used in case a sub-expression mutates the context, but we need to fix it so that subexpressions don't mutate the context: side-effect free, right?
        res.append([ f(ShiftContext(con, obj)) for f in sq_funcs ])
    con.current = orig
    return res


def Filter(con, sourceList, *filters):
    """
    Apply a boolean expression to each of the resources
    The first argument must return a list of resources
    The second is a list of boolean-expressions.
    the return value is the list of filtered resources
    """
    sq_funcs = [
        Optimizer.SubQueryFunction(
          Compile(DataTypes.ToString(f)), con
        )
        for f in filters
        ]
    sourceList = DataTypes.ToList(sourceList)
    res = []
    orig = con.current
    for obj in sourceList:
        #FIXME: ShiftContext is used in case a sub-expression mutates the context, but we need to fix it so that subexpressions don't mutate the context: side-effect free, right?
        if len([ 1 for f in sq_funcs
                   if DataTypes.ToBoolean(f(ShiftContext(con, obj))) ]) == \
                   len(sq_funcs):
            res.append(obj)
    con.current = orig
    return res


def Sort(con, sourceList=None, type_=Constants.SORT_STRING,
         order=Constants.SORT_ASCENDING):
    """
    Sort the given set or list.
    If there are no arguments, the context is sorted using default params
    The first argument is converted to a list
    The second argument is the optional sort type
    The third argument is the optional sort order
    The results are the sorted list
    """
    if sourceList is None:
        sourceList = DataTypes.ToList(con.current)
    else:
        sourceList = DataTypes.ToList(sourceList)
    #Decorate/Sort/Undecorate pattern
    #Decorate
    if DataTypes.Cmp(Constants.SORT_NUMBER, type_) == 0:
        sourceList = [ (i, DataTypes.ToNumber(i)) for i in sourceList ]
    else:
        sourceList = [ (i, DataTypes.ToString(i)) for i in sourceList ]
    #Sort
    sourceList.sort()
    if not DataTypes.Cmp(Constants.SORT_DESCENDING, order):
        sourceList.reverse()
    #Undecorate
    return [ value for i, value in sourceList ]


def SortQ(con, sourceList, expr, type_=Constants.SORT_STRING,
          order=Constants.SORT_ASCENDING):
    """
    Dynamically sort using the given set or list.
    The first argument is converted to a list
    The second argument is the expression to be dynamically evaluated per list item
    The third argument is the optional sort type
    The fourth argument is the optional sort order
    The results are the sorted list
    """
    if sourceList is None:
        sourceList = con.current
    else:
        sourceList = DataTypes.ToList(sourceList)
    expr = Compile(DataTypes.ToString(expr))
    sort_func = lambda a, b: cmp(a[1], b[1])
    key_func = Optimizer.SubQueryFunction(expr, con)
    if DataTypes.Cmp(Constants.SORT_NUMBER, type_) == 0:
        keyList = [ (i, DataTypes.ToNumber(key_func(ShiftContext(con, i)))) for i in sourceList ]
    else:
        keyList = [ (i, DataTypes.ToString(key_func(ShiftContext(con, i)))) for i in sourceList ]
    keyList.sort(sort_func)
    #Remember that the context value is corrupted at this point
    if DataTypes.Cmp(Constants.SORT_DESCENDING, order) == 0:
        keyList.reverse()
    return [ i[0] for i in keyList ]


def ShiftContext(con, new):
    con.current = new
    return con

    
def All(con):
    """
    Get all resources in the system
    returns a list of strings (resources)
    Note, if this was all()[] Then a FilterFunction is created
    """
    return [ DataTypes.ToResource(r) for r in con.driver.resources(con.scope) ]


def Type(con, types):
    """
    Get all resources of a specific type.  This takes into account
    rdfs:subClassOf
    """
    #types = DataTypes.ToList(DataTypes.ToSet(types))
    types = DataTypes.ToList(types)
    if not types:
        return DataTypes.Set()
    full_types = types[:]
    sc_info = ResourceExpressions.GetRelations(None, RDF_SUBCLASSOF_PROP, con, 1)
    for t in types:
        full_types.extend(sc_info.get(t, []))
    #FIXME: The following line can be a performance pig
    type_info = ResourceExpressions.GetRelations(full_types, RDF_TYPE_PROP, con, 1)
    results = []
    for k in type_info.keys():
        results.extend(type_info[k])
    results = DataTypes.Set([ DataTypes.ToResource(r) for r in results ])
    return results


def Intersection(con, *args):
    """
    Set intersection
    """
    set_args = [ DataTypes.ToSet(e) for e in args ]
    res = DataTypes.Set()
    #FIXME: optimize by using the shortest
    for i in set_args[0]._d.keys():
        for a in set_args:
            if i in a._d.keys():
                res.append(i)
    return res


def Union(con, *args):
    """
    Set union
    """
    set_args = [ DataTypes.ToSet(e) for e in args ]
    res = DataTypes.Set()
    #FIXME: optimize by using the shortest
    for a in set_args:
        for i in a._d.keys():
            res.append(i)
    return res


#
# Resource functions
#

def Traverse(con, subjs, preds, dirn=Constants.TRAVERSE_FWD,
             trans=Constants.TRAVERSE_SIMPLE):
    preds = DataTypes.ToList(DataTypes.ToSet(preds))
    start_nodes = DataTypes.ToList(DataTypes.ToSet(subjs))
    results = [ ResourceExpressions.GetRelations(
            start_nodes, p, con, (dirn == Constants.TRAVERSE_INV),
            forceTransitive=(trans == Constants.TRAVERSE_TRANS)
            ) for p in preds ]

    nodes = DataTypes.Set()
    #FIXME: Outer loop over the smaller of start_nodes and sdict.keys()
    for n in start_nodes:
        for r in results:
            if r.has_key(n):
                for obj in r[n]:
                    nodes.append(obj)
    return nodes


def Order(con, subjs, pred, dirn=Constants.TRAVERSE_FWD):
    pred = DataTypes.ToResource(pred)
    relation = ResourceExpressions.GetRelations(
            None, pred, con, (dirn == Constants.TRAVERSE_INV),
            forceTransitive=0
            )
    results = []
    start_nodes = DataTypes.ToList(DataTypes.ToSet(subjs))
    for n in start_nodes:
        subresult = []
        _depthFirst(n, relation, subresult, {})
        results.append(subresult)
    return results


def _depthFirst(curr, relation, subresult, seen):
    #subresult and seen are mutated in place
    subresult.append(curr)
    if seen.has_key(curr):
        #To prevent infinite loops
        return
    seen[curr] = None
    if relation.has_key(curr):
        arcs = relation[curr]
        #arcs = DataTypes.ToSet(relation[curr])
        for next in arcs:
            _depthFirst(next, relation, subresult, seen)
    return
    

def Triples(con, matchSubj, matchPred, matchObj):
    """
    Return a list of triples matching the given criteria
    """
    raise Exception("Not implemented")


#
# Conversions
#

def Resource(con, obj):
    r = DataTypes.ToResource(obj)
    #FIXME: Is this still needed?
    if con.driver.isResource(str(r)):
        return r
    return r


def String(con, obj):
    s = DataTypes.ToString(obj)
    return s


def Number(con, obj):
    s = DataTypes.ToNumber(obj)
    return s


def Boolean(con, obj):
    b = DataTypes.ToBoolean(obj)
    return b


#def List(con, obj):
#    s = DataTypes.ToList(obj)
#    return s


#def Set(con, obj):
#    s = DataTypes.ToSet(obj)
#    return s


def Current(con):
    """
    Return the current value, if there is one
    """
    con.current


#
# Boolean functions
#

def And(con, *args):
    """
    Logical conjunction
    """
    computed = [ arg for arg in args if DataTypes.ToBoolean(arg) ]
    return len(computed) == len(args) and boolean.true or boolean.false


def Or(con, *args):
    """
    Logical disjunction
    """
    computed = [ arg for arg in args if DataTypes.ToBoolean(arg) ]
    return computed and boolean.true or boolean.false


def Not(con, arg):
    """
    Logical negation
    """
    if DataTypes.ToBoolean(arg):
        return boolean.false
    else:
        return boolean.true


def IsResource(con, arg):
    """
    Return true if the item is of type resource, or can be
    successfully converted to resource
    """
    if DataTypes.IsList(arg):
        arg = arg[0]
    if DataTypes.IsSet(arg):
        arg = arg._d[0]
    if DataTypes.IsResource(arg):
        return boolean.true
    if DataTypes.IsString(arg):
        return con.driver.isResource(arg) and boolean.true or boolean.false
    return boolean.false


def Contains(con, arg1, arg2=None):
    """
    contains(outer, inner)
    Return true if inner is contained in outer.
    """
    if arg2 is None:
        inner = arg1
        outer = con.current
    else:
        outer = arg1
        inner = arg2
    if DataTypes.IsString(outer):
        inner = DataTypes.ToString(inner)
        return outer.find(inner) != -1 and boolean.true or boolean.false
    if DataTypes.IsResource(outer):
        outer = DataTypes.ToString(outer)
        inner = DataTypes.ToString(inner)
        return outer.find(inner) != -1 and boolean.true or boolean.false
    return boolean.false


def ContainsCi(con, arg1, arg2=None):
    """
    contains(outer, inner)
    Return true if inner is contained in outer.
    Case insensitive variation
    """
    if arg2 is None:
        inner = arg1
        outer = con.current
    else:
        outer = arg1
        inner = arg2
    if DataTypes.IsString(outer):
        inner = DataTypes.ToString(inner).lower()
        return outer.lower().find(inner) != -1 and boolean.true or boolean.false
    if DataTypes.IsResource(outer):
        outer = DataTypes.ToString(outer).lower()
        inner = DataTypes.ToString(inner).lower()
        return outer.find(inner) != -1 and boolean.true or boolean.false
    return boolean.false


def StartsWith(con, arg1, arg2=None):
    """
    starts-with(outer, inner)
    Return true if outer starts with inner.
    """
    if arg2 is None:
        inner = arg1
        outer = con.current
    else:
        outer = arg1
        inner = arg2
    if DataTypes.IsString(outer):
        inner = DataTypes.ToString(inner)
        return outer.find(inner) == 0 and boolean.true or boolean.false
    if DataTypes.IsResource(outer):
        outer = DataTypes.ToString(outer)
        inner = DataTypes.ToString(inner)
        return outer.find(inner) == 0 and boolean.true or boolean.false
    return boolean.false


def Member(con, arg1, arg2=None):
    """
    member(outer, inner)
    Return true if inner is a member of outer.
    """
    if arg2 is None:
        inner = arg1
        outer = con.current
    else:
        outer = arg1
        inner = arg2
    if DataTypes.IsList(outer):
        return [ i for i in outer if DataTypes.Cmp(i, inner) == 0 ] and boolean.true or boolean.false
    if DataTypes.IsSet(outer):
        return outer._d.has_key(inner) and boolean.true or boolean.false
    return boolean.false


def IsType(con, obj, type_):
    """
    If obj resolves to a list, return true if any member is of the types
    If obj is not a list, return true if the item is of one of the specifed types
    """
    res = DataTypes.ToList(obj)
    #Turn the type into a list of possible types (taking into account SubClass
    #FIXME: also take into account sameClassAs/samePropertyAs
    t = DataTypes.ToList(type_)
    found = DataTypes.Set()
    for i in t:
        found.append(i)
        _expandSubTypes(con.driver, i, found)

    for cur in res:
        for cur_t in found._d:
            if con.driver.complete(cur, _TYPE, cur_t, None, con.scope, {}):
                return boolean.true
    return boolean.false


def _expandSubTypes(driver, baseType, found):
    """
    Find any sub classes of this type
    s = None
    p = rdfs:subClassOf
    o = baseType
    """
    subTypes = driver.subjectsFromPredAndObjs(_SUBCLASSOF, [baseType])
    for s in subTypes:
        if not s in found._d:
            found.append(s)
            #See if it has any sub types
            _expandSubTypes(driver, s, found)


def IsLiteral(con, obj):
    """
    If computed argument is a list or set, convert to list
    and return true if the first member
    is a literal, otherwise return true if arg is a literal
    """
    if DataTypes.IsList(obj):
        obj = obj[0]
    if DataTypes.IsSet(obj):
        obj = obj._d[0]
    if DataTypes.IsResource(obj):
        return boolean.false
    if DataTypes.IsString(obj):
        return not con.driver.isResource(obj) and boolean.true or boolean.false
    return boolean.true


#
# Relational functions
#

# See http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52549
# This does slow down these functions, but to an extent that is drowned out by the general slowness of function dispatch in Versa
class curry:
    def __init__(self, fun, *args, **kwargs):
        self.fun = fun
        self.pending = args[:]
        self.kwargs = kwargs.copy()

    def __call__(self, *args, **kwargs):
        if kwargs and self.kwargs:
            kw = self.kwargs.copy()
            kw.update(kwargs)
        else:
            kw = kwargs or self.kwargs

        return self.fun(*(self.pending + args), **kw)


def RelFunc(con, left, right=None, cmp_results=None):
    #return result of comparison by data types ordering
    #cmp_results is a list of comparison results (0, 1 or 2) that
    #determines what sort of comparison to undertake
    #for example [1] would check if left is bigger than right and [0] whether they are equal.
    if right is None:
        right = left
        left = con.current
    return DataTypes.Cmp(left, right) in cmp_results and boolean.true or boolean.false


Gt = curry(RelFunc, cmp_results=[1])
Lt = curry(RelFunc, cmp_results=[-1])
Gte = curry(RelFunc, cmp_results=[0, 1])
Lte = curry(RelFunc, cmp_results=[0, -1])
Eq = curry(RelFunc, cmp_results=[0])


def Evaluate(con, expr):
    """
    Dynamically evaluate the given string
    """
    expr = Compile(DataTypes.ToString(expr))
    return expr.evaluate(con)


def Properties(con, res=None):
    if res is None: res = con.current
    stmts = con.driver.complete(
        str(DataTypes.ToResource(res)),
        None, None, None, con.scope, {}
        )
    return DataTypes.Set([ DataTypes.ToResource(s[1]) for s in stmts ])

#
# String functions
#

def Concat(con, *args):
    """
    Concatenate a list of strings
    """
    return ''.join([ DataTypes.ToString(s) for s in args ])


def StringLength(con, obj):
    """
    Return the length of a string
    """
    return len(DataTypes.ToString(obj))


def SubstringBefore(con, arg1, arg2=None):
    if arg2 is None:
        inner = arg1
        outer = con.current
    else:
        outer = arg1
        inner = arg2
    index = outer.find(inner)
    if index == -1:
        return ""
    return outer[:index]


def SubstringAfter(con, arg1, arg2=None):
    if arg2 is None:
        inner = arg1
        outer = con.current
    else:
        outer = arg1
        inner = arg2
    index = outer.find(inner)
    if index == -1:
        return ""
    return outer[index+len(inner):]



FUNCTIONS = {
   (None, 'distribute'): Distribute,
   (None, 'traverse'): Traverse,
   (None, 'filter'): Filter,
   (None, 'all'): All,
   (None, 'sort'): Sort,
   (None, 'sortq'): SortQ,
   (None, 'type'): Type,
   (None, 'resource'): Resource,
   (None, 'string'): String,
   (None, 'number'): Number,
   (None, 'list'): List,
   (None, 'set'): Set,
   (None, 'join'): Join,
   (None, 'boolean'): Boolean,
   (None, 'and'): And,
   (None, 'or'): Or,
   (None, 'not'): Not,
   (None, 'is-resource'): IsResource,
   (None, 'contains'): Contains,
   (None, 'contains-ci'): ContainsCi,
   (None, 'starts-with'): StartsWith,
   (None, 'member'): Member,
   (None, 'is-type'): IsType,
   (None, 'is-literal'): IsLiteral,
   (None, 'gt'): Gt,
   (None, 'lt'): Lt,
   (None, 'eq'): Eq,
   (None, 'gte'): Gte,
   (None, 'lte'): Lte,
   (None, 'ge'): Gte,
   (None, 'le'): Lte,
   (None, 'evaluate'): Evaluate,
   (None, 'properties'): Properties,
   (None, 'current'): Current,
   (None, 'intersection'): Intersection,
   (None, 'union'): Union,
   (None, 'substring-after'): SubstringAfter,
   (None, 'substring-before'): SubstringBefore,
   (None, 'concat'): Concat,
   (None, 'string-length'): StringLength,
   (None, 'length'): Length,
   (None, 'triples'): Triples,
   (None, 'head'): Head,
   (None, 'tail'): Tail,
   (None, 'rest'): Rest,
   (None, 'item'): Item,
   (None, 'order'): Order,
    }

