########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Server/Common/Install/InstallUtil.py,v 1.53 2003/04/15 15:34:15 uogbuji Exp $
"""
Definitions and functions used in the installation process

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

import sys, os, sha, urllib, stat
from Ft.Server.Common import Schema, AclConstants
from Ft.Server.Common.ResourceTypes import ResourceType
from Ft.Lib import Time, ProgressIndicator
from distutils import file_util

from Ft.Xml import XPath, Domlette, EMPTY_NAMESPACE
from Ft.Xml.Xslt import XSL_NAMESPACE
from Ft.Lib import Uri


class StatusFlags:
    UNKNOWN = -1
    NO_CHANGE = 0
    NOT_PRESENT = 1
    CONTENT_CHANGE = 2
    IMT_CHANGE = 4
    ACL_CHANGE = 8
    OWNER_CHANGE = 16
    OTHER_CHANGE = 32


    ACTION_REQUIRED = NOT_PRESENT | CONTENT_CHANGE | IMT_CHANGE | ACL_CHANGE | OWNER_CHANGE | OTHER_CHANGE


g_safeAclIdents = [AclConstants.SUPER_USER_GROUP_NAME,
                   AclConstants.WORLD_GROUP_NAME,
                   AclConstants.USERS_GROUP_NAME,
                   AclConstants.ANONYMOUS_USER_NAME,
                   AclConstants.OWNER]


class IndicatorBase:
    indicator = None
    def _message(self,msg,quiet):
        if not quiet:
            if self.indicator:
                self.indicator.message(msg)
            else:
                print msg

    def _setIndicator(self,indicator):
        """
        Store information about where we are in the indicator list
        """
        self.indicator = indicator


class RawFile (IndicatorBase):
    resourceType = ResourceType.RAW_FILE

    contentUri = None
    stringContent = None
    stringLastModified = None  #specified in iso 8601
    lmds = None
    owners = {}
    imts = {}

    def __init__(self, path, acl, owner, imt):
        self.path = path
	self.absPath = path + ';no-traverse'

        #Cache any changes to ACL
        self._aclToUpdate = {}
        self._aclToRemove = []

        self.owner = owner
        self.imt = imt or 'text/plain'

        self.status = StatusFlags.UNKNOWN
        self.dependencies = []
        if acl is None:
            acl = []
        self._setAcl(acl)
        return

    def _setAcl(self, acl):
        """
        Helper function for test harnesses only
        """
        self.acl = acl
        self.aclDict = {}
        for access, ident, allowed in acl:
            if not self.aclDict.has_key(access):
                self.aclDict[access] = {}
            self.aclDict[access][ident] = allowed
        return

    def setPath(self, contentUri):
        self.contentUri = contentUri
        return

    def setSource(self, stringContent, stringLastModified):
        self.stringContent = stringContent
        self.stringLastModified = stringLastModified
        return

    def checkStatus(self, repo, force=0,checkAcl = 1):
        """
        Get the status of this resource from against the supplied repo,
        Note, force here specifies that we force update to the status.  Not force install
        """

        if not force and self.status != StatusFlags.UNKNOWN:
            return self.status

        self._initRdfCaches(repo)

        if not self.lmds.has_key(self.path):
            self.status = StatusFlags.ACTION_REQUIRED

            #Force the ACL changes
            self._aclToUpdate = self.aclDict

            #Exit here
            return self.status
        self.status = StatusFlags.NO_CHANGE

        # Compare the file times.
        localTime = None
        if self.contentUri is not None:
            urlfile = urllib.urlopen(self.contentUri)
            try:
                mtime = os.fstat(urlfile.fileno())[stat.ST_MTIME]
            except OSError:
                # Window's socket streams cannot be stat'ed
                # Get the last modified date from the headers, if possible
                modified = urlfile.info().getheader('Last-Modified')
                if modified:
                    datetime = Time.FromRFC822(modified)
                else:
                    # Unable to determine last modified date, use current
                    # time to force an update
                    datetime = Time.FromPythonTime()
            else:
                datetime = Time.FromPythonTime(mtime)

            urlfile.close()
            localTime = datetime.asISO8601DateTime()
        elif self.stringLastModified is not None:
            localTime = self.stringLastModified

        if localTime is not None:
            # Compare it with the repo.  Some things (i.e., containers) don't
            # have a modified date check.
            if localTime > self.lmds[self.path]:
                self.status = self.status | StatusFlags.CONTENT_CHANGE

        # Check for other kinds of changes
	if checkAcl:
	    self._checkAclChange(repo)
        self._checkImtChange(repo)
        self._checkOwnerChange(repo)
        self._checkOtherChange(repo)

        return self.status

    def createContent(self, repo, quiet):
        """
        Called when we need to add our content to the repo
        """
        self._message("Create RawFile: %s IMT=%s"%(self.path, self.imt),quiet)
        repo.createRawFile(self.path, self.imt, self._getContent())
        #Don't need to reset this on a create
        self.status = self.status & ~StatusFlags.IMT_CHANGE
        return

    def updateContent(self, repo, quiet):
        """
        Called when we need to update our content in the repo
        """
        self._message("Modify %s content"%self.path,quiet)
        repo.fetchResource(self.path).setContent(self._getContent())
        return

    def updateImt(self, repo, quiet):
        """
        Called when we need to update our imt in the repo
        """
        self._message("Setting IMT of %s --> %s" % (self.path, self.imt),quiet)
        repo.fetchResource(self.absPath).setImt(self.imt)
        return

    def updateOwner(self, repo, quiet):
        """
        Called when we need to update our owner in the repo
        """
        self._message("Setting Owner of %s --> %s" % (self.path, self.owner),quiet)
        owner = repo.fetchUserOrGroupByName(self.owner)
        #Owner will be None if 'world'
        if owner:
            repo.fetchResource(self.absPath).setOwner(owner)
        return

    def updateAcl(self, repo, quiet):
        """
        Called when we need to update our acl in the repo
        """
        res = repo.fetchResource(self.absPath)
        for access in self._aclToRemove:
            self._message("Removing ACL: %s from %s" % (access, self.path),quiet)
            res.inheritAcl(access)

        for access,info in self._aclToUpdate.items():
            self._message("Setting ACL: %s for %s" % (access, self.path),quiet)
            l = info.items()
            for ident,allowed in l[:1]:
                res.setAcl(access, ident, allowed)
            for ident,allowed in l[1:]:
                res.addAcl(access, ident, allowed)
        return

    def updateOther(self, repo, quiet):
        """
        Called when we need to update our other stuff in the repo
        """
        pass

    def _getContent(self):
        if self.stringContent:
            return self.stringContent
        elif self.contentUri:
            content = urllib.urlopen(self.contentUri).read()
            return content
        else:
            raise RuntimeError("Attempting to update %s with no content" % self.path)
        return

    def _checkAclChange(self, repo):
        """
        Look for changes in the ACL
        """
        res = repo.fetchResource(self.absPath)

        cur = res.getAcl()

        for access,info in self.aclDict.items():
            if not cur.has_key(access):
                #Add the entire access
                self._aclToUpdate[access] = info
            else:
                #Verify all of the keys are the same
                for ident,allowed in info.items():
                    if cur[access].get(ident,-1) != allowed:
                        #The access is the same, but the ident/allowed is different
                        #Add the whole thing
                        self._aclToUpdate[access] = info
                        break
                else:
                    #Make sure that cur doesn't define any extra idents
                    for ident,allowed in cur[access].items():
                        if not self.aclDict[access].has_key(ident):
                            #It does so rewrite the entire access
                            self._aclToUpdate[access] = info
                            break

        #Lastly, make sure that cur does not define any other access
        for access in cur.keys():
            if not self.aclDict.has_key(access):
                self._aclToRemove.append(access)

        if len(self._aclToUpdate.keys()) or len(self._aclToRemove):
            self.status = self.status | StatusFlags.ACL_CHANGE

        return

    def _checkImtChange(self,repo):
        """
        Look for changes in the IMT
        """
        if self.imt and self.imt != self.imts[self.path]:
            self.status = self.status | StatusFlags.IMT_CHANGE
        return

    def _checkOwnerChange(self,repo):
        """
        Look for changes in the Owner
        """
        if self.owner and self.owner != self.owners[self.path]:
            self.status = self.status | StatusFlags.OWNER_CHANGE
        return

    def _checkOtherChange(self,repo):
        """
        Look for other changes
        """
        pass

    def _setBase(self, baseDir):
        """
        Adjust for a relative installation
        """
        self.path = Uri.BaseJoin(baseDir, self.path)
	self.absPath = self.path + ';no-traverse'
        return

    def _gatherDependencies(self,baseDir):
        """
        Get a list of dependencies for this resource.  At the RawFile level, it is dependencies
        based on path
        """
        self.dependencies = []
        base = os.path.dirname(self.path)
        if not (base == baseDir or base+'/' == baseDir):
            self.dependencies.append(base)
        return

    def _initRdfCaches(self,repo):
        """
        We use RDF completes to get a lot of the information about the resources.  This speeds up installation a lot.
        """
        if self.lmds is not None: return
        RawFile.lmds = {}
        for lmd in repo.getModel().complete(None, Schema.MODIFIED_DATE, None):
            self.lmds[lmd.subject] = lmd.object

        for imt in repo.getModel().complete(None, Schema.IMT, None):
            self.imts[imt.subject] = imt.object

        for owner in repo.getModel().complete(None, Schema.OWNER, None):
            self.owners[owner.subject] = owner.object
        return

    def __repr__(self):
        return "<%s to be installed at %s id=%s>" % (self.__class__.__name__,
                                                     self.path,
                                                     str(hex(id(self))))
class UriReferenceFile(RawFile):
    resourceType = ResourceType.URI_REFERENCE_FILE

    _refs = None

    def __init__(self,path,acl,owner,imt,href):
        self.href = href
        RawFile.__init__(self,path,acl,owner,imt)
        return

    def _checkOtherChange(self,repo):
        """
        See if the doc def has changed
        """
        RawFile._checkOtherChange(self,repo)
        if self.href != self._refs[self.path]:
            self.status = self.status | StatusFlags.OTHER_CHANGE
        return

    def createContent(self,repo,quiet):
        """
        Called when we need to add our content to the repo
        """
        self._message("Create UriReferenceFile: %s IMT=%s reference=%s"%(self.path, self.imt, self.href),quiet)
        repo.createUriReferenceFile(self.path, self.imt, self.href)
        #Don't need to reset this on a create
        self.status = self.status & ~StatusFlags.IMT_CHANGE
        return

    def updateContent(self,repo,quiet):
        """
        Called when we need to update our content in the repo
        """
        raise RuntimeError("Unable to update content for %s" % self.path)

    def _initRdfCaches(self,repo):
        RawFile._initRdfCaches(self,repo)

        if self._refs is not None: return
        UriReferenceFile._refs = {}
        for ref in repo.getModel().complete(None, Schema.URI_REFERENCE_LOCATION, None):
            self._refs[ref.subject] = ref.object
        return


class XmlDocument(RawFile):
    resourceType = ResourceType.XML_DOCUMENT

    _dds = None

    docName='XmlDocument'

    def __init__(self, path, acl, owner, imt, docDef):
        RawFile.__init__(self, path, acl, owner, imt)
        self.docDef = docDef
        if imt is None:  self.imt = 'text/xml'
        return

    def createContent(self, repo, quiet):
        """
        Called when we need to add our content to the repo
        """
        self._message("Create %s: %s IMT=%s%s" % (self.docName,
                       self.path, self.imt,
                       " docDef=" * (not not self.docDef) + (self.docDef or "")),
                      quiet)

        cnt = self._getContent()
        repo.createDocument(self.path, cnt, docDef = self.docDef,
                            imt=self.imt, forcedType = self.resourceType)

        #Don't need to reset this on a create
        self.status = self.status & ~StatusFlags.IMT_CHANGE

        #Note, we need to be smart in updateOther cause we already set the doc def
        return

    def updateOther(self, repo, quiet):
        RawFile.updateOther(self, repo, quiet)
        doc = repo.fetchResource(self.absPath)
        curDD = doc.getDocumentDefinition()
        if (curDD is None and self.docDef is not None) or (curDD is not None and curDD.getAbsolutePath() != self.docDef):

            self._message("Setting %s document definition to %s" % (self.path, self.docDef),quiet)
            dd = None
            if self.docDef is not None:
                dd = doc.fetchResource(self.docDef)
            doc.setDocumentDefinition(dd)
        return

    def _setBase(self, baseDir):
        """
        Adjust for a relative installation
        """
        RawFile._setBase(self, baseDir)
        if self.docDef is not None:
            self.docDef = Uri.BaseJoin(baseDir, self.docDef)
        return

    def _checkOtherChange(self, repo):
        """
        See if the doc def has changed
        """
        RawFile._checkOtherChange(self,repo)
        if self.docDef != self._dds[self.path]:
            self.status = self.status | StatusFlags.OTHER_CHANGE
        return

    def _gatherDependencies(self,baseDir):
        RawFile._gatherDependencies(self, baseDir)
        #Find dependencies from resolution needed during parsing
        #FIXME: Only works for one resolution dependency at a time
        reader = Domlette.NonvalidatingReader
        self.cachedDoc = None
        try:
            if self.contentUri:
                doc = reader.parseUri(self.contentUri)
            elif self.stringContent:
                doc = reader.parseString(self.stringContent)
            else:
                doc = None
            self.cachedDoc = doc
        except Uri.UriException, e:
            self.dependencies.append(e.params[0])
        except:
            pass
        if self.docDef is not None:
            self.dependencies.append(self.docDef)
        return

    def _initRdfCaches(self,repo):
        RawFile._initRdfCaches(self,repo)

        if self._dds is not None: return
        XmlDocument._dds = {}
        for dd in repo.getModel().complete(None, Schema.DOCDEF, None):
            if dd.object == Schema.NULL_DOCDEF:
                self._dds[dd.subject] = None
            else:
                self._dds[dd.subject] = dd.object
        return


class Container(XmlDocument):
    resourceType = ResourceType.CONTAINER

    def __init__(self, path, acl, owner, imt, docDef):
        if path[-1] == '/' and len(path) > 1: path = path[:-1]
        XmlDocument.__init__(self, path, acl, owner, imt, docDef)
        return

    def createContent(self, repo, quiet):
        self._message("Create Container: %s" % self.path,quiet)
        repo.createContainer(self.path, 1)
        if self.imt == 'text/xml':
            self.status = self.status & ~StatusFlags.IMT_CHANGE
        return

    def updateContent(self, repo, quiet):
        """
        Called when we need to update our content in the repo
        """
        raise RuntimeError("Unable to update content for %s" % self.path)


import os
from Ft import SHARE_DIR
EXT_DIR = os.path.join(SHARE_DIR, 'ExtensionModules')
from distutils.dep_util import newer

class ExtensionModule(IndicatorBase):
    resourceType = None
    owner=-1
    def __init__(self, file):
        if not file:
            raise Exception('File attribute not given')
        self._file = Uri.UriToOsPath(file)
        self._dest = os.path.join(EXT_DIR, os.path.basename(self._file))
        self.path = ''
        self.moduleName = "Ft.Share.ExtensionModules." + os.path.splitext(os.path.basename(self._file))[0]
        self.status = StatusFlags.UNKNOWN
        return

    def createContent(self, repo, quiet):
        #repo.copyFile(open(self._file, 'r').read(), )
        res = file_util.copy_file(self._file,
                                  self._dest,
                                  update=1,
                                  )
        if res[1]:
            self._message("Copy extension module: %s" % self._file,quiet)
        return

    def updateContent(self, repo, quiet):
        self.createContent(repo, quiet)
        repo.reloadModule(self.moduleName)

    def checkStatus(self, repo, force=0,checkAcl = 1):
        self.status = StatusFlags.NO_CHANGE
        if os.path.exists(self._dest):
            if force or newer(self._file, self._dest):
                self.status = StatusFlags.CONTENT_CHANGE
        else:
            self.status = StatusFlags.NOT_PRESENT
        return self.status

    def _setBase(self, baseDir): return

    def _gatherDependencies(self, baseDir): self.dependencies = []

    def setPath(self,fName): pass

class XsltDocument(XmlDocument):
    resourceType = ResourceType.XSLT_DOCUMENT
    dependencyXPath = XPath.Compile('/xsl:*/xsl:import | /xsl:*/xsl:include')
    docName = 'XsltDocument'

    def _gatherDependencies(self, baseDir):
        """
        The dependencies of a Xslt document are all xslt:include and xslt:import elements
        """
        XmlDocument._gatherDependencies(self, baseDir)
        if not self.cachedDoc: return
        con = XPath.Context.Context(self.cachedDoc,
                                    processorNss={'xsl': XSL_NAMESPACE})

        for dep in self.dependencyXPath.evaluate(con):
            d = dep.getAttributeNS(EMPTY_NAMESPACE, 'href')
            if d:
                if ':' in d:
                    #External to the system
                    continue
                if d[0] != '/':
                    d = Uri.BaseJoin(self.path, d)
                self.dependencies.append(d)
        return


class SchematronDocument(XmlDocument):
    resourceType = ResourceType.SCHEMATRON_DOCUMENT
    docName='SchematronDocument'

class RdfDocument(XmlDocument):
    resourceType = ResourceType.RDF_DOCUMENT
    docName='RdfDocument'

class XsltDocumentDefinition(XmlDocument):
    resourceType = ResourceType.XSLT_DOCUMENT_DEFINITION
    docName='XsltDocumentDefinition'

class XPathDocumentDefinition(XmlDocument):
    resourceType = ResourceType.XPATH_DOCUMENT_DEFINITION
    docName='XPathDocumentDefinition'

class Server(XmlDocument):
    resourceType = ResourceType.SERVER
    docName='Server'

class Command(XmlDocument):
    resourceType = ResourceType.COMMAND
    docName='Command'

class Alias(XmlDocument):
    resourceType = ResourceType.ALIAS
    _refs = None

    def __init__(self, path, acl, owner, imt, docDef, reference):
        self.reference = reference
        XmlDocument.__init__(self,path,acl,owner,imt,docDef)

        if self.reference[0] != '/':
            #Make it absolute
            self.reference = Uri.BaseJoin(self.path, self.reference)
        return

    def createContent(self,repo,quiet):
        """
        Called when we need to add our content to the repo
        """
        self._message("Create Alias: %s"%self.path,quiet)

        doc = repo.fetchResource(self.reference)
        doc.addAlias(self.path)

        if self.imt == 'text/xml':
            self.status = self.status & ~StatusFlags.IMT_CHANGE

        #Note, we need to be smart in updateOther cause we already set the reference
        return

    def updateContent(self, repo, quiet):
        """
        Called when we need to update our content in the repo
        """
        raise RuntimeError("Unable to update content for %s" % self.path)

    def updateOther(self,repo,quiet):
        XmlDocument.updateOther(self,repo,quiet)

        #You cannot change an aliases reference, so raise exception for now, we could lose ACL, owner, etc
        if repo.fetchResource(self.absPath).getReference().getAbsolutePath() != self.reference:
            raise RuntimeError("Unable to change the alias reference for %s, please remove alias by hand.  You can do so on the command line using '4ss delete --no-traverse %s'" % (self.path, self.path))
        return

    def _checkOtherChange(self,repo):
        """
        See if the doc def has changed
        """
        XmlDocument._checkOtherChange(self,repo)
        if self.reference != self._refs[self.path]:
            self.status = self.status | StatusFlags.OTHER_CHANGE
        return

    def _setBase(self,baseDir):
        """
        Adjust for a relative installation
        """
        XmlDocument._setBase(self,baseDir)
        self.reference = Uri.BaseJoin(self.path, self.reference)
        return

    def _gatherDependencies(self,baseDir):
        XmlDocument._gatherDependencies(self,baseDir)
        self.dependencies.append(self.reference)
        return

    def _initRdfCaches(self,repo):
        XmlDocument._initRdfCaches(self,repo)

        if self._refs is not None: return
        Alias._refs = {}
        for ref in repo.getModel().complete(None, Schema.ALIAS_REFERENCE, None):
            self._refs[ref.subject] = ref.object
        return


class User(XmlDocument):
    resourceType = ResourceType.USER

    def __init__(self, userName, acl, owner, imt, docDef, basePath,
                 password=None, data={}, passwdHash=None):
        XmlDocument.__init__(self, basePath + '/' + userName, acl, owner,
                             imt, docDef)
        self.userName = userName
        self.basePath = basePath

        if passwdHash:
            self.passwordHash = passwdHash
            self.password = password
        else:
            self.passwordHash = sha.new(password).hexdigest()
            self.password = password

        self.data = data
        return

    def createContent(self, repo, quiet):
        """
        Called when we need to add our content to the repo
        """
        self._message("Create User: %s%s" % (self.path,
                       " docDef=" * (not not self.docDef) + (self.docDef or "")),
                      quiet)

        newUser=repo.fetchResource(self.basePath).createUser(self.userName,
                                                             self.passwordHash,
                                                             self.docDef)
        newUser.setUserData(self.data)

        #Note, we need to be smart in updateOther cause we already set the password
        if self.imt == 'text/xml':
            self.status = self.status & ~StatusFlags.IMT_CHANGE
        return

    def updateContent(self, repo, quiet):
        """
        Called when we need to update our content in the repo
        """
        raise RuntimeError("Unable to update content for %s" % self.path)

    def updateOther(self, repo, quiet):
        XmlDocument.updateOther(self, repo, quiet)

        if repo.fetchResource(self.path).getPassword() != self.passwordHash:
            self._message("Updating User %s password" % self.userName,quiet)
            repo.fetchResource(self.path).setPassword(self.passwordHash)

        if repo.fetchResource(self.path).getUserData() != self.data:
            self._message("Updating User %s's data" % self.userName,quiet)
            repo.fetchResource(self.path).setUserData(self.data)
        return

    def _checkAclChange(self, repo):

        #Users get thier ACL set automatically by the system.  We need to account for that here.
        #So, we need to update the ACL list before we do the check, unless an ACL list was specified

        if not self.acl:
            for key in [AclConstants.DELETE_ACCESS,
                        AclConstants.CHANGE_PERMISSIONS_ACCESS,
                        AclConstants.WRITE_USER_MODEL_ACCESS,
                        AclConstants.CHANGE_OWNER_ACCESS,
                        AclConstants.READ_ACCESS,
                        AclConstants.EXECUTE_ACCESS,
                        AclConstants.WRITE_ACCESS]:
                for ident in [self.userName,
                              AclConstants.OWNER,
                              repo.getUserName()]:
                    self.acl.append((key,ident,AclConstants.ALLOWED))
        return XmlDocument._checkAclChange(self,repo)

    def _checkOtherChange(self,repo):
        """
        See if the doc def has changed
        """
        XmlDocument._checkOtherChange(self,repo)
        if self.passwordHash != repo.fetchResource(self.path).getPassword():
            self.status = self.status | StatusFlags.OTHER_CHANGE
        return

    def _setBase(self, baseDir):
        """
        Adjust for a relative installation
        """
        XmlDocument._setBase(self, baseDir)
        self.basePath = Uri.BaseJoin(baseDir, self.basePath)
        return


class Group(XmlDocument):
    resourceType = ResourceType.GROUP

    def __init__(self, groupName, acl, owner, imt, docDef, basePath, members):
        XmlDocument.__init__(self, basePath + '/' + groupName, acl, owner,
                             imt, docDef)

        if not groupName:
            raise SyntaxError("Must specify a group name in a member definition")
        self.groupName = groupName
        self.members = members
        self.basePath = basePath

        self._membersToAdd = []
        self._membersToRemove = []
        return

    def checkStatus(self,repo,force=0,checkAcl = 1):
        rt = XmlDocument.checkStatus(self,repo,force,checkAcl)

        ## Should this be 'self.status & StatusFlags.ACTION_REQUIRED'??
        if self.status & StatusFlags.NOT_PRESENT:
            self._membersToAdd = self.members
        return rt

    def createContent(self,repo,quiet):
        """
        Called when we need to add our content to the repo
        """
        self._message("Create Group: %s"%self.path,quiet)
        self._message("creating members: %s" % self.members,quiet)
        g = repo.fetchResource(self.basePath).createGroup(self.groupName)
        if self.imt == 'text/xml':
            self.status = self.status & ~StatusFlags.IMT_CHANGE
        return

    def updateContent(self,repo,quiet):
        """
        Called when we need to update our content in the repo
        """
        raise RuntimeError("Unable to update content for %s" % self.path)

    def updateOther(self, repo, quiet):
        XmlDocument.updateOther(self,repo,quiet)

        g = repo.fetchUserOrGroupByName(self.groupName)
        for m in self._membersToAdd:
            self._message("Adding member %s to group %s" % (m,self.groupName),quiet)
            mem = repo.fetchUserOrGroupByName(m)
            g.addMember(mem)
        for m in self._membersToRemove:
            self._message("Removing member %s from group %s" % (m,self.groupName),quiet)
            mem = repo.fetchUserOrGroupByName(m)
            g.removeMember(mem)
        return

    def _checkAclChange(self, repo):

        #Groups get thier ACL set automatically by the system.  We need to account for that here.
        #So, we need to update the ACL list before we do the check, unless an ACL list was specified

        if not self.acl:
            for key in [AclConstants.DELETE_ACCESS,
                        AclConstants.CHANGE_PERMISSIONS_ACCESS,
                        AclConstants.WRITE_USER_MODEL_ACCESS,
                        AclConstants.CHANGE_OWNER_ACCESS,
                        AclConstants.READ_ACCESS,
                        AclConstants.EXECUTE_ACCESS,
                        AclConstants.WRITE_ACCESS]:
                for ident in [self.groupName,
                              AclConstants.OWNER,
                              repo.getUserName()]:
                    self.acl.append((key, ident, AclConstants.ALLOWED))
        return XmlDocument._checkAclChange(self, repo)

    def _checkOtherChange(self, repo):
        """
        See if the doc def has changed
        """
        XmlDocument._checkOtherChange(self, repo)

        curMembers = map(lambda x:x.getUsername(),repo.fetchUserOrGroupByName(self.groupName).getMembers())
        for m in curMembers:
            if m not in self.members:
                self._membersToRemove.append(m)
        for m in self.members:
            if m not in curMembers:
                self._membersToAdd.append(m)
        if len(self._membersToRemove) or len(self._membersToAdd):
            self.status = self.status | StatusFlags.OTHER_CHANGE
        return

    def _setBase(self, baseDir):
        """
        Adjust for a relative installation
        """
        XmlDocument._setBase(self, baseDir)
        self.basePath = Uri.BaseJoin(baseDir, self.basePath)
        return

    def _gatherDependencies(self, baseDir):
        XmlDocument._gatherDependencies(self, baseDir)
        for m in self.members:
            if m not in g_safeAclIdents:
                self.dependencies.append(m)
        return


class Product:
    name = ""
    version = 0
    description = ""
    useIndicator = 1

    def __init__(self, resourceList, name='', version=0, description=''):
        self.resourceList = resourceList
        self.filteredList = None
        self.sortedList = None
        self.baseDir = '/'
        self.name = name
        self.version = version
        self.description = description

        return

    def setBase(self, baseDir):
        """
        Walk the list of resources, and resolve them to a new base directory
        """
        if baseDir[-1] != '/':
            baseDir = baseDir + '/'
        if baseDir[0] != '/':
            baseDir = '/' + baseDir

        for r in self.resourceList:
            r._setBase(baseDir)
        self.baseDir = baseDir
        return

    def setIndicator(self):
        """
        Walk the list of resources, and resolve them to a new base directory
        """

        for r in self.resourceList:
            r._setIndicator(self.progressIndicator)
        return

    def filter(self, repo,checkAcl = 1):
        #The filter list will filter all of the resources.
        #It splits them into two lists.  The first is all of them that
        #need to have thier content and/or doc defs updated, the second are ones that
        #just need properties updated.  The second list does not need to have its
        #dependencies expanded.

        if self.filteredList is not None:
            return

        self.filteredList = []
        self.sortedList = []

        for r in self.resourceList:
            status = r.checkStatus(repo,checkAcl = checkAcl)
            if status == StatusFlags.UNKNOWN:
                raise RuntimeError("Status of %s not determined" % r.path)
            if status & (StatusFlags.CONTENT_CHANGE | StatusFlags.NOT_PRESENT):
                self.sortedList.append(r)
            elif status != StatusFlags.NO_CHANGE:
                self.filteredList.append(r)
        return

    def sort(self, repo):
        # Sort our resources by dependency

        # We assume that the resources have been filtered, so first gather
        # dependencies on the sortedList
        for r in self.sortedList:
            r._gatherDependencies(self.baseDir)

        sorted = []
        provides = []  # list of resource paths
        unsorted = self.sortedList

        # Transfer from the unsorted list to the res list.  If we every don't
        # Transfer any, then blindly move one over and print an error
        satisfied = 1
        while unsorted and satisfied:
            # Start off by assuming nothing will be satified this round
            satisfied = []

            # Find all resources whose dependencies are satisfied
            for resource in unsorted:
                for path in resource.dependencies:
                    if path not in provides:
                        if repo.hasResource(path):
                            provides.append(path)
                        else:
                            # Unknown
                            break
                else:
                    # All dependencies satisfied
                    satisfied.append(resource)
                    provides.append(resource.path)
                    if resource.resourceType == ResourceType.USER:
                        provides.append(resource.userName)
                    elif resource.resourceType == ResourceType.GROUP:
                        provides.append(resource.groupName)

            for resource in satisfied:
                sorted.append(resource)
                unsorted.remove(resource)

        # Gather the paths which have failed for the remaining resources
        if unsorted:
            failed = []
            for resource in unsorted:
                for path in resource.dependencies:
                    if path not in provides:
                        failed.append((path, resource.path))

            failed.sort()
            sys.stderr.write("Failed dependencies:\n")
            for item in failed:
                sys.stderr.write("\t%s needed by %s\n" % item)

        self.sortedList = sorted
        return not unsorted

    def serialize(self, stream=sys.stdout, refUri=''):
        import Serialize
        stream.write(Serialize.Serialize(self, refUri))
        return

    def install(self, repo, baseDir, quiet, do_update=0,checkAcl = 1):
        """
        Install the product into the repo
        """
        action = ('Installing','Updating')[do_update]
        header = ""
        if self.version:
            header = "\n%s %s version %s" % (action, self.name, self.version)
        else:
            header = "\n%s %s" % (action, self.name)
        print header

        self.setBase(baseDir)
        self.filter(repo,checkAcl = checkAcl)

        if not self.sort(repo):
            raise RuntimeError("Unresolved dependencies exist")


        if self.useIndicator:
            #Calculate the total size
            #We go over the resource list twice, filter, sort, header
            total = len(self.sortedList) + len(self.sortedList) + len(self.filteredList)
            if total:
                self.progressIndicator = ProgressIndicator.ProgressIndicator(self.name)
                self.progressIndicator.newIndicator(total)
                self.setIndicator()

        ctr = 0
        #First install the sorted list
        for r in self.sortedList:
            if r.status & StatusFlags.NOT_PRESENT:
                r.createContent(repo, quiet)
                #Might need to remove owner update if not specified
                if r.owner is None:
                    r.status = r.status & ~StatusFlags.OWNER_CHANGE

            else:
                r.updateContent(repo,quiet)
            r.status = r.status & ~StatusFlags.CONTENT_CHANGE

            if self.useIndicator:
                self.progressIndicator.updateProgress(ctr)
            ctr += 1

        #Then the filtered list
        for r in self.filteredList + self.sortedList:
            if r.status & StatusFlags.IMT_CHANGE:
                r.updateImt(repo, quiet)
                r.status = r.status & ~StatusFlags.IMT_CHANGE
            if r.status & StatusFlags.ACL_CHANGE:
                r.updateAcl(repo, quiet)
                r.status = r.status & ~StatusFlags.ACL_CHANGE
            if r.status & StatusFlags.OWNER_CHANGE:
                r.updateOwner(repo, quiet)
                r.status = r.status & ~StatusFlags.OWNER_CHANGE
            if r.status & StatusFlags.OTHER_CHANGE:
                r.updateOther(repo, quiet)
                r.status = r.status & ~StatusFlags.OTHER_CHANGE
            if self.useIndicator:
                self.progressIndicator.updateProgress(ctr)
            ctr += 1
        return


def Deserialize(stream, refUri=''):
    import Serialize
    return Serialize.Deserialize(stream, refUri)

