########################################################################
#
# File Name:            HeadDict.py
#
"""
Provides an implementation of header storage for HTTP.

Copyright 2002 Fourthought, Inc. (USA)
See http://4suite.org/COPYRIGHT for license and copyright information.
"""

import UserDict

class HeaderDict(UserDict.UserDict):
    """
    A dictionary-like object for storing MIME-like message headers.
    Headers are stored in the object via regular dictionary interfaces.
    They are retrieved via str(), repr() or self.terse(), which returns
    a standard dict with duplicate headers merged into a comma-separated
    list of values (even when it is not OK to do so).

    The reason this class is needed is because the dictionary interface of
    the Message class of the rfc822 and mimetools modules allows each new
    header to overwrite a previously-stored header with the same key. Our
    HeaderDict class stores its header values as lists, so no data is lost.
    Its keys are accessed case-insensitively and are always written out
    in a normalized format where usually only the first letter of each
    hyphen-separated word is capitalized. (Content-MD5, ETag, TE, and
    WWW-Authenticate are exceptions to this rule). The str(obj)
    representation prints the headers in a convenient format suitable for
    printing in a MIME-like message.
    """

    # exceptions to the automatic camel-capping
    forceCaps = { 'content-md5': 'Content-MD5',
                  'etag': 'ETag',
                  'te': 'TE',
                  'www-authenticate': 'WWW-Authenticate'
                }

    # camel-cap a hyphenated string
    def _camelCap(self, s, checkForceCaps=0):
        if len(s) == 0:
            return ''
        if checkForceCaps:
            if s.lower() in forceCaps.keys():
                return forceCaps[s.lower()]
        return '-'.join(map(lambda s: s[0].upper() + s.lower()[1:], s.split('-')))

    # case-insensitive lookup
    def __getitem__(self, key):
        return self.data[key.lower()]

    # case-insensitive storage; appends to a list
    def __setitem__(self, key, value):
        if self.has_key(key):
            self.data[key.lower()].append(value)
        else:
            self.data[key.lower()] = [value]

    # case-insensitive deletion: removes entire list
    def __delitem__(self, key):
        del self.data[key.lower()]

    # retrieve a value or default if no such key
    def get(self, key, default=None):
        if self.has_key(key):
            return self.__getitem__(key)
        else:
            return default

    # case-insensitive key lookup
    def has_key(self, key):
        return self.data.has_key(key.lower())

    # normalize key names
    def keys(self):
        return [self._camelCap(key, 0) for key in self.data.keys()]

    # return a terse dictionary (duplicate headers in one comma-separated string)
    # this is for compatibility with functions want to read the header values as
    # a single string rather than as a python list of strings
    def terse(self):
        dict = {}
        for key in self.keys():
            dict[key] = ', '.join(self.data[key.lower()])
        return dict

    # append another dict or HeaderDict
    def update(self, dict):
        if type(dict) == type({}):
            for key in dict.keys():
                self.__setitem__(key, dict[key])
        elif type(dict) == type(self):
            for key in dict.keys():
                for val in dict[key]:
                    self.__setitem__(key, val)
        else:
            raise("HeaderDict.update() called with a non-dictionary argument")

    # string representations
    def __repr__(self):
        # dictionary syntax
        return '{%s}' % ', '.join([("%s: %s" % (repr(key), repr(self.data[key.lower()]))) for key in self.keys()])
    def __str__(self):
        # a series of 'Hyphenated-Header: value' CRLF-terminated lines, or empty string
        if len(self.keys()):
            return '\r\n'.join(['\r\n'.join(i) for i in [["%s: %s" % (key, val) for val in self.data[key.lower()]] for key in self.keys()] ]) + '\r\n'
        return ''
