########################################################################
# $Source: /var/local/cvsroot/4Suite/Ft/Xml/Catalog.py,v $ $Revision: 1.10 $ $Date: 2003/04/18 20:41:07 $
"""
OASIS TR9401 and XML Input Resolver.  Based on contribution from Tarn Weisner Burton to PyXML.  See https://sourceforge.net/tracker/index.php?func=detail&aid=490069&group_id=6473&atid=306473

Copyright 2002 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

'''
A catalog resolver that implements partial support for XML Catalogs and TR9401 Catalogs.

This module contains the following classes and functions:

Catalog - a class which implements the catalog resolver and parser including support for
          xml.sax.handler.EntityResolver and URI resolution with the method resolveURI.

read_catalogs - a function which reads a collection of catalogs from an environment
                variable list.

Author:  Tarn Weisner Burton <twburton@users.sf.net>
'''

from Ft.Xml import InputSource, READ_EXTERNAL_DTD
import os, re, string, warnings
from cStringIO import StringIO
import urlparse, urllib
import xml.sax, xml.sax.handler
from Ft import SHARE_DIR
from Ft.Lib import Uri

TR9401 = re.compile(r'^\s*(BASE|CATALOG|DELEGATE|PUBLIC|SYSTEM|OVERRIDE\s+YES|OVERRIDE\s+NO)\s+"((?:[^"\\]|\\.)*)"(?:\s+"((?:[^"\\]|\\.)*)")?', re.M | re.I)
hex_code = re.compile('%(..)')


class CatalogInputSource(InputSource.InputSource):
    def __init__(self, catalog, *args, **kwargs):
        apply(InputSource.InputSource.__init__, (self,) + args, kwargs)
        self.catalog = catalog
        self.fallbackResolve = InputSource.InputSource.resolve
        return

    def resolve(self, systemId, publicId=None, hint=None, ignoreErrors=0):
        '''resolve an entity from pubid if possible, otherwise use systemId'''
        # first unwrap the URNs
        unwrapped, publicId = unwrap_urn(publicId)
        unwrapped, systemId = unwrap_urn(systemId)
        # if the systemId was really a public URN then replace publicId
        if unwrapped:
            if not publicId:
                publicId = systemId
            elif publicId != systemId and not self.quiet:
                print "Error: publicId doesn't match unwrapped systemId"
            systemId = None

        # try to do the actual resolution
        result = self.__resolveEntity(publicId, systemId)

        # if the resolution failed then return systemId
        if result == 1 or result is None:
            # revert to base behaviour
            systemId = self.resolveURI(systemId)
            return self.fallbackResolve(self, systemId, publicId, hint, ignoreErrors=ignoreErrors)
        else:
            #Use the resulting sysid
            #Do we cascade this through the URI resolver as well?
            return self.fallbackResolve(self, result, None, hint, ignoreErrors=ignoreErrors)

    def clone(self, stream, uri, hint=None):
        "Override the cloning of this input source to account for the extra param."
        return self.__class__(self.catalog, stream, uri,
                              processIncludes=self.processIncludes,
                              stripElements=self.stripElements,
                              factory=self.factory)

    def __match_partial(self, list, x):
        '''construct a list of items whose keys are a partial match for a given string'''

        matches = []
        for key, value in list:
            if key == x[:len(key)]:
                matches.append((key, value))

        # sort the results with the longest matches first
        matches.sort(lambda a, b: cmp(len(b[0]), len(a[0])))
        return matches

    def __rewrite(self, list, x):
        '''rewrite a string based on partial matches'''

        matches = self.__match_partial(list, x)
        if len(matches):
            # just use the longest one
            return matches[0][1] + x[len(matches[0][0]):]

    def __resolveEntity(self, publicId, systemId):
        '''
        attempt to resolve an entity.  Success is indicated by a returned string,
        failure by None and abort by a 1
        '''

        if systemId is not None:
            # If a system identifier is provided, and at least one matching system entry exists,
            # the (absolutized) value of the uri attribute of the first matching system entry is returned.
            if self.catalog.systemIds.has_key(systemId):
                return self.catalog.systemIds[systemId]

            # If a system identifier is provided, and at least one matching rewriteSystem entry exists,
            # rewriting is performed.
            new_systemId = self.__rewrite(self.catalog.systemRewrites, systemId)
            if new_systemId is not None:
                return new_systemId

            # If a system identifier is provided, and one or more delegateSystem entries match,
            # delegation is performed
            matches = self.__match_partial(self.catalog.systemDelegates, systemId)
            if len(matches):
                for match in matches:
                    result = match.__resolveEntity(publicId, systemId)
                    if result:
                        return result
                # no matches found so abort
                return 1

        if publicId is not None:
            # If a public identifier is provided, and one or more delegatePublic entries match,
            # delegation is performed
            matches = self.__match_partial(self.catalog.publicDelegates, publicId)

            for i in range(len(matches)-1, -1, -1):
                if not matches[i][1] and systemId is not None:
                    del matches[i]

            if len(matches):
                for match in matches:
                    result = match[0].__resolveEntity(publicId, systemId)
                    if result:
                        return result
                # no matches found so abort
                return 1

            # If a public identifier is provided, and at least one matching public entry exists, the
            # (absolutized) value of the uri attribute of the first matching public entry is returned.
            # If a system identifier is also provided as part of the input to this catalog lookup, only
            # public entries that occur where the prefer setting is public are considered for matching.
            if self.catalog.publicIds.has_key(publicId):
                uri, prefer_public = self.catalog.publicIds[publicId]
                if prefer_public or systemId is None:
                    return uri

        # If there are one or more catalog entry files remaining on the current catalog entry file list,
        # load the next catalog entry file and continue resolution efforts
        for catalog in self.catalog.catalogs:
            result = catalog.__resolveEntity(publicId, systemId)
            if result:
                return result

    def __resolveURI(self, uri):
        '''attempt to resolve a URI.  Success is indicated by a returned string,
        failure by None and abort by a 1'''

        # If at least one matching uri entry exists, the (absolutized) value of the uri attribute of
        # the first matching uri entry is returned
        if self.catalog.uris.has_key(uri):
            return self.catalog.uris[uri]

        # If at least one matching rewriteURI entry exists, rewriting is performed.
        #if self.catalog.systemRewrites.has_key(systemId):
        new_uri = self.__rewrite(self.catalog.uriRewrites, uri)
        if new_uri is not None:
            return new_uri

        # If one or more delegateURI entries match, delegation is performed
        matches = self.__match_partial(self.catalog.uriDelegates, uri)
        if len(matches):
            for match in matches:
                result = match.__resolveURI(uri)
                if result:
                    return result
            # no matches found so abort
            return 1

        # If there are one or more catalog entry files remaining on the current catalog entry file list,
        # load the next catalog entry file and continue resolution efforts
        for catalog in self.catalog.catalogs:
            result = catalog.__resolveURI(uri)
            if result:
                return result


    def resolveURI(self, uri):
        '''resolve a URI if possible, return the original URI otherwise'''

        result = self.__resolveURI(uri)

        # no match found return the original
        if result == 1 or result is None:
            return uri
        return result


def urn_normalize(x):
    '''normalize a URN'''

    if x and string.lower(x[:4]) == 'urn:':
        # make the first two components lowercase
        parts = string.split(x, ':', 2)
        parts[0] = string.lower(parts[0])
        parts[1] = string.lower(parts[1])

        # make hex codes uppercase
        parts = hex_code.split(string.join(parts, ':'))
        for i in range(1, len(parts), 2):
            parts[i] = '%' + string.upper(parts[i])
        return string.join(parts, '')

    return x


def unwrap_urn(x):
    '''unwrap a public URN.'''

    x = urn_normalize(x)
    if x and x[:13] == 'urn:publicid:':
        x = string.replace(x[13:], '+', ' ')
        x = string.replace(x, ':', '//')
        x = string.replace(x, ';', '::')
        parts = hex_code.split(x)
        for i in range(1, len(parts), 2):
            parts[i] = chr(int(parts[i], 16))
        x = string.join(parts, '')
        return (1, x)
    return (0, x)


class Catalog(xml.sax.handler.ContentHandler, InputSource.InputSourceFactory):
    '''
    A Catalog resolver that can parse
    OASIS TR 9401 Catalogs <http://www.oasis-open.org/specs/a401.htm>
    and OASIS XML Catalogs <http://www.oasis-open.org/committees/entity/spec.html>
    Can be used as an InputSourceFactory
    '''
    def __init__(self, uri, quiet=1):
        self.systemIds = {}
        self.publicIds = {}
        self.uris = {}
        self.publicDelegates = []
        self.systemDelegates = []
        self.uriDelegates = []
        self.systemRewrites = []
        self.uriRewrites = []
        self.catalogs = []
        self.uri = uri
        self.quiet = quiet

        #Set up an input source for the catalog file itself
        factory = InputSource.DefaultFactory
        self.isrc = factory.fromUri(self.uri)
        #self.isrc = InputSource.InputSource(self.uri, self.uri)
        data = self.isrc.read()
        self.parse(data)
        return

    def fromStream(self, stream, uri=None, *v_args, **kw_args):
        """
        parse the content of the given stream.  uri is the URI or system ID
        to use for the entity of the string
        """
        if not uri:
            warnings.warn("Creation of InputSource without an URI",
                          RuntimeWarning, 2)
        kw_args['factory'] = self
        return apply(CatalogInputSource, (self, stream, uri) + v_args, kw_args)


    def parse(self, data):
        if re.findall('<.*catalog\s', data):
            # it's XML so set the default options and parse the file
            self.parseXmlCat(data)
        else:
            # probably a TR 9401 file
            self.parseTr9401(data)
        return

    def parseXmlCat(self, data):
        """
        As specified in
        http://www.oasis-open.org/committees/entity/spec-2001-08-06.html
        Partially implemented
        """
        self.prefer_public = [1]
        self.base = [self.uri]
        p = xml.sax.make_parser()

        # Set the parser to use the globally defined external entity parsing
        # setting.
        try:
            p.setFeature(xml.sax.handler.feature_external_ges,
                         READ_EXTERNAL_DTD)
        except (xml.sax.SAXNotRecognizedException,
                xml.sax.SAXNotSupportedException):
            pass

        p.setContentHandler(self)
        p.parse(self.uri)
        del self.prefer_public
        del self.base
        return

    def parseTr9401(self, data):
        """
        As specified in
        http://www.oasis-open.org/specs/a401.htm
        Partially implemented
        """
        prefer_public = 1
        base = self.uri
        for cmd in TR9401.findall(data):
            token = string.upper(cmd[0])
            if token == 'PUBLIC':
                if len(cmd) == 3:
                    self.publicIds[cmd[1]] = (urlparse.urljoin(base, cmd[2]), prefer_public)
            elif token == 'SYSTEM':
                if len(cmd) == 3:
                    self.systemIds[cmd[1]] = urlparse.urljoin(base, cmd[2])
            elif token == 'BASE':
                base = cmd[1]
            elif token[:8] == 'OVERRIDE':
                prefer_public = string.strip(token[8:]) == 'YES'
            elif token == 'DELEGATE':
                if len(cmd) == 3:
                    self.publicDelegates[cmd[1]] = urlparse.urljoin(base, cmd[2])
            elif token == 'CATALOG':
                if len(cmd) == 2:
                    self.catalogs.append(Catalog(urlparse.urljoin(base, cmd[1]), self.quiet))
        return

    # methods used by the XML parser

    def __make_abs(self, attrs, attr_name):
        '''make a URI absolute'''

        # look for a xml:base attribute, otherwise use the current base
        if attrs.has_key('xml:base'):
            base = attrs['xml:base']
        else:
            base = self.base[-1]

        return urlparse.urljoin(base, attrs[attr_name])


    def startElement(self, name, attrs):
        '''start an element in the XML parser'''
        if name == 'public':
            # a publicId lookup
            if self.__ensure_attrs(name, attrs, 'publicId', 'uri'):
                # save the state of prefer_public also
                self.publicIds[attrs['publicId']] = (self.__make_abs(attrs, 'uri'), self.prefer_public[-1])
        elif name == 'system':
            # a systemId lookup
            if self.__ensure_attrs(name, attrs, 'systemId', 'uri'):
                self.systemIds[attrs['systemId']] = self.__make_abs(attrs, 'uri')
        elif name == 'uri':
            # a URI lookup
            if self.__ensure_attrs(name, attrs, 'name', 'uri'):
                self.uris[attrs['name']] = self.__make_abs(attrs, 'uri')
        elif name == 'rewriteURI':
            # a URI rewrite
            if self.__ensure_attrs(name, attrs, 'uriStartString', 'rewritePrefix'):
                self.uriRewrites.append((attrs['uriStartString'], self.__make_abs(attrs, 'rewritePrefix')))
        elif name == 'rewriteSystem':
            # a systemId rewrite
            if self.__ensure_attrs(name, attrs, 'systemIdStartString', 'rewritePrefix'):
                self.systemRewrites.append((attrs['systemIdStartString'], self.__make_abs(attrs, 'rewritePrefix')))
        elif name == 'delegateSystem':
            # delegate systemId to specific catalog
            if self.__ensure_attrs(name, attrs, 'systemIdStartString', 'catalog '):
                self.systemRewrites.append((attrs['systemIdStartString'], Catalog(self.__make_abs(attrs, 'catalog'), self.quiet)))
        elif name == 'delegatePublic':
            # delegate publicId to specific catalog
            if self.__ensure_attrs(name, attrs, 'publicIdStartString', 'catalog '):
                # save the state of prefer_public also
                self.systemRewrites.append((attrs['publicIdStartString'], (Catalog(self.__make_abs(attrs, 'catalog'), self.quiet), prefer_public)))
        elif name == 'delegateURI':
            # delegate URI to specific catalog
            if self.__ensure_attrs(name, attrs, 'uriStartString', 'catalog '):
                self.systemRewrites.append((attrs['uriStartString'], Catalog(self.__make_abs(attrs, 'catalog'), self.quiet)))
        elif name == 'nextCatalog':
            # the next catalog in a chain
            if self.__ensure_attrs(name, attrs, 'catalog'):
                self.catalogs.append(Catalog(self.__make_abs(attrs, 'catalog'), self.quiet))
        elif name == 'catalog':
            # look for prefer or xml:base attributes and setup the base and prefer stacks
            if attrs.has_key('prefer'):
                self.prefer_public = [attrs['prefer'] == 'public']
            if attrs.has_key('xml:base'):
                self.base = [attrs['xml:base']]
        elif name == 'group':
            # look for prefer or xml:base attributes and push them onto the right stack
            if attrs.has_key('prefer'):
                self.prefer_public.append(attrs['prefer'] == 'public')
            else:
                self.prefer_public.append(self.prefer_public[-1])

            if attrs.has_key('xml:base'):
                self.base.append(attrs['xml:base'])
            else:
                self.base.append(self.base[-1])


    def __ensure_attrs(self, name, attrs, *attr_names):
        '''ensure that the right attributes exist just in case the parser is a non-validating one'''

        for attr_name in attr_names:
            if not attrs.has_key(attr_name):
                if not self.quiet:
                    print '%s: Malformed %s element, missing %s attribute' % (self.uri, name, attr_name)
                return 0
        return 1


    def endElement(self, name):
        '''end an element in the XML parser'''

        if name == 'group':
            # pop the stacks
            self.prefer_public.pop()
            self.base.pop()


    def startElementNS(self, name, qname, attrs):
        '''start an XML element with namespace support'''

        # defer to the ordinary method
        self.startElement(name[1], attrs)


    def endElementNS(self, name, qname):
        '''end an XML element with namespace support'''

        # defer to the ordinary method
        self.endElement(name[1])


def read_catalogs(environ_name='XML_CATALOGS', quiet=1):
    '''read catalogs from a list in an environment variable'''

    catalog = None

    if os.environ.has_key('XML_CATALOGS'):
        for uri in string.split(os.environ['XML_CATALOGS'], os.pathsep):
            if catalog:
                catalog.catalogs.append(Catalog(uri, quiet))
            else:
                catalog = Catalog(uri, quiet = quiet)

    return catalog


if os.environ.has_key("XML_CATALOGS"):
    FT_CATALOG = read_catalogs()
else:
    try:
        FT_CATALOG = Catalog(Uri.OsPathToUri(os.path.join(SHARE_DIR, "default.cat")))
    except Exception, e:
        raise e
        import sys
        sys.stderr.write("Default catalog %s not found\n"%os.path.join(SHARE_DIR, "default.cat"))
        FT_CATALOG = None


class CatalogInputSourceFactory(InputSource.InputSourceFactory):
    """
    A factory for creating new InputSource instances that use a given catalog.
    """
    def __init__(self, inputSourceClass=None, resolver=Uri.BASIC_RESOLVER,
                 catalog=FT_CATALOG):
        self._klass = inputSourceClass or CatalogInputSource
        self.resolver = resolver
        self.catalog = catalog
        return

    def fromStream(self, stream, uri=None, *v_args, **kw_args):
        """
        Creates an InputSource from the given stream.
        The uri argument is the URI to use for the stream
        (one should always be given, even if it's bogus).
        """
        if not uri:
            warnings.warn("Creation of InputSource without a URI",
                          RuntimeWarning, 2)
        kw_args['factory'] = self
        if not kw_args.has_key('resolver'): kw_args['resolver'] = self.resolver
        return apply(self._klass, (self.catalog, stream, uri)+v_args, kw_args)

DefaultCatalogInputSourceFactory = CatalogInputSourceFactory()


if __name__ == '__main__':
    import sys, getopt

    print 'Reading catalogs...'
    c = read_catalogs(quiet = 0)
    if c is None: c = FT_CATALOG
    cSource = CatalogInputSource(c, StringIO('<dummy/>'), 'file:///dummy')
    print

    options, argv = getopt.getopt(sys.argv[1:], 'e:u:', ('entity=', 'uri='))
    for option, value in options:
        if option == '-e' or option == '--entity':
            publicId, systemId = string.split(value, ',')
            if publicId == '':
                publicId = None
            if systemId == '':
                systemId = None
            print 'Resolve entity...'
            print '\tpublicId = "%s"' % publicId
            print '\tsystemId = "%s"' % systemId
            print '\tresult = "%s"' % cSource._CatalogInputSource__resolveEntity(publicId, systemId)
            print
        elif option == '-u' or option == '--uri':
            print 'Resolve URI...'
            print '\turi = "%s"' % value
            print '\tresult = "%s"' % cSource._CatalogInputSource__resolveURI(value)
            print

