########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Xml/InputSource.py,v 1.25 2003/03/18 08:03:51 uogbuji Exp $
"""
Classes providing a standard interface and encapsulation of metadata for
document/entity streams intended for input to various XML processors.

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

import cStringIO, types, warnings

from Ft.Lib import Uri, Uuid


class InputSource:
    """
    An input source is an encapsulation of a source of content.
    It includes a stream (Python file-like object) from which the
    content can be read, a URI to identify the stream and facilitate
    resolution of relative URI references / system IDs encountered
    within the stream, and parameters used by the processors of the
    stream (XML parsers, XSLT processors).

    It is designed to be overridden as applications need different
    functionality from sources.
    """
    def __init__(self, stream, uri=None, processIncludes=1, stripElements=None,
                 factory=None, resolver=Uri.BASIC_RESOLVER):
        """
        stream - the stream associated with this input source
        uri - the absolute URI of the input source
        processIncludes=1 - Whether or not XIncludes should be expanded
        stripElements - Space stripping rules
        factory - The factory that created this instance
        """
        if uri:
            self.uri = uri
        else:
            self.uri = 'urn:uuid:'+Uuid.UuidAsString(Uuid.GenerateUuid())
        self.stream = stream
        self.processIncludes = processIncludes
        self.stripElements = stripElements or []
        self.factory = factory
        frag_split = self.uri.split("#")
        self.fragment = ""
        if len(frag_split) == 2:
            self.fragment = frag_split[1]
        self._resolver = resolver
        return

    def resolve(self, sysid, pubid=None, hint=None, ignoreErrors=0):
        """
        Resolves a system identifier (fragmentless URI reference) or a
        public identifier into a new input source. It is typically used
        when a URI reference is encountered in the original stream and
        needs to be resolved (e.g. to support an external entity
        reference, XInclude, xsl:include/import, document(), etc.).

        The hint parameter is used to give a hint as to what the
        resolution will be used for.

        The default implementation only
        resolves system identifiers; the pubid and hint are ignored.

        If the ignoreErrors flag is set, an error during resolution
        (such as "file not found") will result in None being returned,
        rather than raising an exception.
        """
        uri = self._normalize(sysid)
        stream = self._openStream(uri, ignoreErrors)
        return self.clone(stream, uri)

    def getUriResolver(self):
        """
        This method returns the URI resolver that is used by this
        input source to normalize (resolve to absolute form) and
        resolve (dereference) URI references. This is the public method
        to use if just URI resolution is needed.
        """
        return self._resolver

    #Stream protocol methods
    def read(self, bufsiz=-1):
        return self.stream.read(bufsiz)

    def readline(self):
        return self.stream.readline()

    def close(self):
        return self.stream.close()

    #Helper Methods
    def _normalize(self, uri):
        """
        Normalize (resolve to absolute form) a given URI reference,
        using the URI of this input source as the base.

        The default implementation will just use the default URI resolver.

        If your input source is working with non-standard or not supported
        URIs, then you will need to override this or the getUriResolver method.
        """
        return self.getUriResolver().normalize(uri, self.uri)

    def _openStream(self, uri, ignoreErrors=0):
        """
        Returns a representation of a resource as a stream by
        resolving the given URI. If ignoreErrors is set, failure to
        obtain the stream will result in None being returned, rather
        than an exception (e.g. "file not found") being raised.

        Default behaviour is to use the resolver associated with this
        InputSource. If your custom InputSource needs to open URIs
        that are not supported natively by this InputSource (e.g.,
        repository objects, or objects from a database), then you
        should override this method and do whatever it takes to
        resolve the URI into a readable stream.
        """
        try:
            return self.getUriResolver().resolve(uri)
        except:
            if ignoreErrors:
                return None
            raise

    def clone(self, stream, uri=None, hint=None):
        """
        Clones this input source, creating a new instance with
        the known params.

        If your derived InputSource requires additional state information
        then you have to override how it is cloned and pickled.
        """
        if uri == None: uri = self.uri
        if stream is None: return NullInputSource(uri)
        return self.__class__(stream, uri,
                              processIncludes=self.processIncludes,
                              stripElements=self.stripElements,
                              factory=self.factory, resolver=self._resolver)

    #Pickle routines.  you need to be able to pickle an input source
    #We cannot pickle a stream though.
    def __getstate__(self):
        state = self.__dict__.copy()
        state['stream'] = None
        return state


class NullInputSource(InputSource):
    """
    An InputSource that simulates an empty stream.
    """
    def __init__(self, uri=None):
        InputSource.__init__(self, None, uri)

    def resolve(self, sysid, pubid=None, hint=None, ignoreErrors=0):
        stream = self._openStream(sysid, ignoreErrors)
        return self.clone(stream, sysid)

    def read(self, bufsiz=-1):
        return ''

    def close(self):
        return


class InputSourceFactory:
    """
    A factory for creating new InputSource instances.
    """
    def __init__(self, inputSourceClass=None, resolver=Uri.BASIC_RESOLVER):
        self._klass = inputSourceClass or InputSource
        self.resolver = resolver
        return

    def fromUri(self, uri, *v_args, **kw_args):
        """
        Creates an InputSource from the stream resulting from the
        resolution of the given URI.

        uri - a URI from which the input will be read.  Important: a file
              path is generally not a URI to be safe, if you wish to read
              from a file use the following pattern:
              from Ft.Lib import Uri
              uri = Uri.OsPathToUri("/path/to/file.ext")
              OR uri = Uri.OsPathToUri("C:\\path\\to\\file.ext")
        """
        src = apply(self.fromStream, (None, uri) + v_args, kw_args)
        stream = src._openStream(src.uri)
        return src.clone(stream, src.uri)

    def fromString(self, st, uri=None, *v_args, **kw_args):
        """
        Creates an InputSource from a stream derived from the given
        string. The uri argument is the URI to use for the stream
        (one should always be given, even if it's bogus).
        """
        if not isinstance(st, types.StringType):
            raise ValueError("String must be of type string, not %s" % type(st))
        stream = cStringIO.StringIO(st)
        return apply(self.fromStream, (stream, uri) + v_args, kw_args)

    def fromStream(self, stream, uri=None, *v_args, **kw_args):
        """
        Creates an InputSource from the given stream.
        The uri argument is the URI to use for the stream
        (one should always be given, even if it's bogus).
        """
        if not uri:
            warnings.warn("Creation of InputSource without a URI",
                          RuntimeWarning, 2)
        kw_args['factory'] = self
        if not kw_args.has_key('resolver'): kw_args['resolver'] = self.resolver
        return apply(self._klass, (stream, uri)+v_args, kw_args)


DefaultFactory = InputSourceFactory()
