########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Xml/Lib/HtmlPrettyPrinter.py,v 1.8 2003/02/23 09:05:02 mbrown Exp $
"""
This module supports formatted document serialization in HTML syntax.

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

from HtmlPrinter import HtmlPrinter

class HtmlPrettyPrinter(HtmlPrinter):
    """
    An HtmlPrettyPrinter instance provides functions for serializing an
    XML or XML-like document to a stream, based on SAX-like event calls
    initiated by an Ft.Xml.Lib.Print.PrintVisitor instance.

    The methods in this subclass of HtmlPrinter attempt to emit a
    document conformant to the HTML 4.01 syntax, with extra whitespace
    added for visual formatting. The indent attribute is the string used
    for each level of indenting. It defaults to 2 spaces.
    """

    # The amount of indent for each level of nesting
    indent = '  '

    def __init__(self, stream, encoding):
        HtmlPrinter.__init__(self, stream, encoding)
        self._level = 0

        # indenting control variables
        self._isInline = [1]  # prevent newline before first element
        self._indentForbidden = 0
        self._inNoIndent = [0]
        self._elementStarted = None
        self._justPrintedLF = 0
        return

    def startElement(self, tagName, namespaces, attributes):
        self._justPrintedLF = 0
        if not self._isInline[-1] and not self._indentForbidden:
            self.writeAscii('\n' + (self.indent * self._level))

        HtmlPrinter.startElement(self, tagName, namespaces, attributes)

        # Setup indenting rules for this element
        name = tagName.lower()
        self._isInline.append(self.inlineElements.has_key(name))
        self._inNoIndent.append(self.noIndentElements.has_key(name))
        self._indentForbidden += self._inNoIndent[-1]
        self._level += 1
        self._elementStarted = tagName
        return

    def endElement(self, tagName):
        self._justPrintedLF = 0
        # Undo changes to indenting rules for this element
        # was_empty means was either empty or had only 1 line of text content
        was_empty = self._elementStarted and self._elementStarted == tagName
        self._elementStarted = None
        self._level -= 1
        inline = self._isInline.pop()
        no_indent = self._inNoIndent.pop()
        self._indentForbidden -= no_indent
        if not inline and not no_indent and not was_empty:
            self.writeAscii('\n' + (self.indent * self._level))
        HtmlPrinter.endElement(self, tagName)
        return

    def text(self, data, disableEscaping=0):
        if data:
            if self._indentForbidden:
                HtmlPrinter.text(self, data, disableEscaping)
            else:
                lines = data.split('\n')
                if len(lines) > 1:
                    self._elementStarted = None
                line = lines[0]
                if self._justPrintedLF:
                    # \n was last character written, so first line doesn't need one
                    self.writeAscii((len(line) > 0) * (self.indent * self._level))
                HtmlPrinter.text(self, line, disableEscaping)
                for line in lines[1:]:
                    self.writeAscii('\n' + (len(line) > 0) * (self.indent * self._level))
                    HtmlPrinter.text(self, line, disableEscaping)
            self._justPrintedLF = (data[-1] == '\n')
        return

    def processingInstruction(self, target, data):
        self._justPrintedLF = 0
        self._elementStarted = None
        # try to indent
        if not self._isInline[-1] and not self._indentForbidden:
            self.writeAscii('\n' + (self.indent * self._level))
        HtmlPrinter.processingInstruction(self, target, data)
        return

    def comment(self, data):
        self._justPrintedLF = 0
        self._elementStarted = None
        # try to indent
        if not self._isInline[-1] and not self._indentForbidden:
            self.writeAscii('\n' + (self.indent * self._level))
        HtmlPrinter.comment(self, data)
        return


    # Elements that should never be emitted on a new line.
    # <td> is added because Netscape 4.x treats whitespace
    # between <tr> and <td> as significant in some situations.
    inlineElements = {}
    for name in ['tt', 'i', 'b', 'u', 's', 'strike', 'big', 'small', 'em',
                 'strong', 'dfn', 'code', 'samp', 'kbd', 'var', 'cite',
                 'abbr', 'acronym', 'a', 'img', 'applet', 'object', 'font',
                 'basefont', 'script', 'map', 'q', 'sub', 'sup', 'span',
                 'bdo', 'iframe', 'input', 'select', 'textarea', 'label',
                 'button', 'td']:
        inlineElements[name] = 1

    # Elements that should never be emitted with additional
    # whitespace in their content; i.e., once you're inside
    # one, you don't do any more indenting.
    noIndentElements = {}
    for name in ['script', 'style', 'pre', 'textarea', 'xmp']:
        noIndentElements[name] = 1

    del name
